@@ -20,6 +20,10 @@
  *
  */
 
+#define COMMAND_OPTION_USERNAME		"username"
+#define COMMAND_OPTION_PASSWORD		"password"
+#define COMMAND_OPTION_PASSPHRASE	"passphrase"
+
 typedef char *(*command_completion_func_t) (const char *text, int state);
 
 enum cmd_status {
@@ -49,8 +53,12 @@ struct command_family {
 	command_completion_func_t family_arg_completion;
 	command_completion_func_t entity_arg_completion;
 	void (*set_default_entity)(const char *entity);
+	void (*reset_default_entity)(void);
 };
 
+const char *command_option_get(const char *name);
+bool command_has_options(void);
+
 bool command_line_find_token(const char *token, uint8_t num_to_inspect);
 char **command_completion(const char *text, int start, int end);
 char *command_entity_arg_completion(const char *text, int state,
@@ -62,6 +70,7 @@ void command_noninteractive_trigger(void);
 bool command_is_interactive_mode(void);
 int command_get_exit_status(void);
 void command_set_exit_status(int status);
+void command_reset_default_entities(void);
 
 void command_family_register(const struct command_family *family);
 void command_family_unregister(const struct command_family *family);
